
//
// This source file is part of appleseed.
// Visit http://appleseedhq.net/ for additional information and resources.
//
// This software is released under the MIT license.
//
// Copyright (c) 2017 Francois Beaune, The appleseedhq Organization
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

// appleseed.renderer headers.
#include "renderer/api/object.h"
#include "renderer/api/project.h"
#include "renderer/api/rendering.h"
#include "renderer/api/scene.h"
#include "renderer/api/types.h"

// todo: fix.
#include "renderer/kernel/shading/shadingray.h"

// appleseed.foundation headers.
#include "foundation/math/ray.h"
#include "foundation/math/scalar.h"
#include "foundation/math/vector.h"
#include "foundation/utility/api/specializedapiarrays.h"
#include "foundation/utility/containers/dictionary.h"
#include "foundation/utility/job/iabortswitch.h"
#include "foundation/utility/searchpaths.h"
#include "foundation/utility/string.h"

// appleseed.main headers.
#include "main/dllvisibility.h"

// BRL-CAD standard headers.
#include <algorithm>
#include <cmath>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdio.h>

// BRL-CAD headers
#include "common.h"
#include "vmath.h"
#include "raytrace.h"

namespace asf = foundation;
namespace asr = renderer;

namespace
{
    //
    // A sphere object.
    //
    // The sphere is assumed to be centered at the origin.
    //

    const char* Model = "brlcad_object";
    bool hit = false;


    int brlcad_hit(struct application *ap, struct partition *PartHeadp, struct seg *UNUSED(segs))
    {
        /* TODO: Add logic */
        hit = true;
        return 1;
    }


    int brlcad_miss(struct application *UNUSED(ap))
    {
        /* TODO: Add logic */
        hit = false;
        return 0;
    }


    class BrlcadObject
      : public asr::ProceduralObject
    {
      public:
        // Constructor.
        BrlcadObject(
            const char*                 name,
            const asr::ParamArray&      params)
          : asr::ProceduralObject(name, params)
          , m_lazy_region_kit(&m_region_kit)
        {
        }

        // Delete this instance.
        void release() override
        {
            rt_clean_resource_complete(0, ap->a_resource);
            bu_free(ap->a_resource, "appleseed");
            bu_free(ap, "appleseed");
            delete this;
        }

        // Return a string identifying this object model.
        const char* get_model() const override
        {
            return Model;
        }

        // This method is called once before rendering each frame.
        // Returns true on success, false otherwise.
        virtual bool on_frame_begin(
            const asr::Project&         project,
            const asr::BaseGroup*       parent,
            asr::OnFrameBeginRecorder&  recorder,
            asf::IAbortSwitch*          abort_switch) override
        {
            if (!asr::ProceduralObject::on_frame_begin(project, parent, recorder, abort_switch))
                return false;

            m_radius = get_uncached_radius();
            m_rcp_radius = 1.0 / m_radius;

            if (!m_prepped_brlcad)
                return prep_brlcad(get_brlcad_database_path(), get_brlcad_object_name());

            return true;
        }

        // Compute the local space bounding box of the object over the shutter interval.
        asr::GAABB3 compute_local_bbox() const override
        {
            const auto r = static_cast<asr::GScalar>(get_uncached_radius());

            if (m_prepped_brlcad) {
                struct rt_i *rtip = ap->a_rt_i;
                if (rtip->needprep)
                    rt_prep_parallel(rtip, 1);
                // double x = (rtip->mdl_min[X] + rtip->mdl_max[X]) / 2;
                // double y = (rtip->mdl_min[Y] + rtip->mdl_max[Y]) / 2;
                // double z = (rtip->mdl_min[Z] + rtip->mdl_max[Z]) / 2;

                // RENDERER_LOG_INFO("Calculated BRL-CAD bounding box: %f %f %f, %f %f %f", -x, -y, -z, x, y, z);

                RENDERER_LOG_INFO("Calculated BRL-CAD bounding box: (%f, %f, %f), (%f, %f, %f)", rtip->mdl_min[X], rtip->mdl_min[Y], rtip->mdl_min[Z], rtip->mdl_max[X], rtip->mdl_max[Y], rtip->mdl_max[Z]);

                // return asr::GAABB3(asr::GVector3(-x, -y, -z), asr::GVector3(x, y, z));
                return asr::GAABB3(asr::GVector3(rtip->mdl_min[X], rtip->mdl_min[Y], rtip->mdl_min[Z]), asr::GVector3(rtip->mdl_max[X], rtip->mdl_max[Y], rtip->mdl_max[Z]));
            }
            return asr::GAABB3(asr::GVector3(-r), asr::GVector3(r));
        }

        // Return the region kit of the object.
        asf::Lazy<asr::RegionKit>& get_region_kit() override
        {
            return m_lazy_region_kit;
        }

        // Access materials slots.
        size_t get_material_slot_count() const override
        {
            return 1;
        }

        const char* get_material_slot(const size_t index) const override
        {
            return "default";
        }

        // Compute the intersection between a ray expressed in object space and
        // the surface of this object and return detailed intersection results.
        void intersect(
            const asr::ShadingRay&  ray,
            IntersectionResult&     result) const override
        {
            const double Epsilon = 1.0e-6;

            const double a = asf::dot(ray.m_org, ray.m_dir);
            const double b = asf::square(a) - dot(ray.m_org, ray.m_org) + asf::square(m_radius);

            if (m_prepped_brlcad) {
                struct rt_i* rtip = ap->a_rt_i;
                if (rtip->needprep)
                    rt_prep(rtip);
                VSET(ap->a_ray.r_pt, ray.m_org[0], ray.m_org[1], ray.m_org[2]);
                VSET(ap->a_ray.r_dir, ray.m_dir[0], ray.m_dir[1], ray.m_dir[2]);

                rt_shootray(ap);
            }            

            // result.m_hit = hit;

            // if (!result.m_hit)
                // return;

            if (b < 0.0)
            {
                result.m_hit = false;
                return;
            }

            const double c = std::sqrt(b);

            double t = -a - c;
            if (t < std::max(ray.m_tmin, Epsilon) || t >= ray.m_tmax)
            {
                t = -a + c;
                if (t < std::max(ray.m_tmin, Epsilon) || t >= ray.m_tmax)
                {
                    result.m_hit = false;
                    return;
                }
            }

            // result.m_hit = true;
            result.m_distance = t;
            // result.m_distance = ap->a_ray_length;

            const asf::Vector3d n = asf::normalize(ray.point_at(t));
            result.m_geometric_normal = n;
            result.m_shading_normal = n;
            // result.m_geometric_normal = partition->pt_inhit->hit_normal;
            // result.m_shading_normal = partition->pt_inhit->hit_normal;

            const asf::Vector3f p(ray.point_at(t) * m_rcp_radius);
            result.m_uv[0] = std::acos(p.y) * asf::RcpPi<float>();
            result.m_uv[1] = std::atan2(-p.z, p.x) * asf::RcpTwoPi<float>();

            result.m_material_slot = 0;
        }

        // Compute the intersection between a ray expressed in object space and
        // the surface of this object and simply return whether there was a hit.
        bool intersect(
            const asr::ShadingRay&  ray) const override
        {
            const double Epsilon = 1.0e-6;

            const double a = asf::dot(ray.m_org, ray.m_dir);
            const double b = asf::square(a) - dot(ray.m_org, ray.m_org) + asf::square(m_radius);

            if (b < 0.0)
                return false;

            const double c = std::sqrt(b);

            const double t1 = -a - c;
            if (t1 >= std::max(ray.m_tmin, Epsilon) && t1 < ray.m_tmax)
                return true;

            const double t2 = -a + c;
            if (t2 >= std::max(ray.m_tmin, Epsilon) && t2 < ray.m_tmax)
                return true;

            return false;
        }

      private:
        asr::RegionKit              m_region_kit;
        asf::Lazy<asr::RegionKit>   m_lazy_region_kit;
        double                      m_radius;
        double                      m_rcp_radius;
        bool                        m_prepped_brlcad = false;

        struct application*         ap;
        struct rt_i*                rtip;

        double get_uncached_radius() const
        {
            return m_params.get_optional<double>("radius");
        }

        std::string get_brlcad_database_path() const
        {
            return m_params.get_required<std::string>("database_path", "");
        }

        std::string get_brlcad_object_name() const
        {
            return m_params.get_required<std::string>("object_name", "");
        }

        bool prep_brlcad(
            std::string  db_path,
            std::string  object_name)
        {
            if (rt_uniresource.re_magic != RESOURCE_MAGIC)
                 rt_init_resource(&rt_uniresource, 0, NULL);

            ap = static_cast<application*>(bu_calloc(1, sizeof(resource), "appleseed"));
            ap->a_resource = static_cast<resource*>(bu_calloc(1, sizeof(resource), "appleseed"));

            char title[1024] = {0};


            // rtip = (rt_i*) malloc(sizeof(rt_i));
            bu_log("dbpath: %s\n", db_path.c_str());
            bu_log("objectpath: %s\n", object_name.c_str());

            rtip = rt_dirbuild(db_path.c_str(), title, sizeof(title));
            if (rtip == RTI_NULL) {
                RENDERER_LOG_ERROR("Failed building the database directory for %s\n", db_path);
                return false;
            }

            if (title[0]) { 
                RENDERER_LOG_INFO("Database title: %s\n", title);
            }

            if (rt_gettree(rtip, object_name.c_str()) < 0) {
                RENDERER_LOG_ERROR("Loading the geometry for [%s] FAILED\n", object_name.c_str());
                return false;
            }
            RENDERER_LOG_INFO("Loaded the geometry for [%s]", object_name.c_str());

            RT_APPLICATION_INIT(ap);
            if (rtip->needprep)
                rt_prep_parallel(rtip, 1);
            // rt_prep(rtip);

            ap->a_rt_i = rtip;
            ap->a_onehit = 0;

            // ap->a_hit = brlcad_hit;
            // ap->a_miss = brlcad_miss;
            // rt_init_resource(&rt_uniresource, 0, ap->a_rt_i);

            m_prepped_brlcad = true;
            return true;
        }
    };


    //
    // Factory for the new object model.
    //

    class BrlcadObjectFactory
      : public asr::IObjectFactory
    {
      public:
        // Delete this instance.
        void release() override
        {
            delete this;
        }

        // Return a string identifying this object model.
        const char* get_model() const override
        {
            return Model;
        }

        // Return metadata for this object model.
        asf::Dictionary get_model_metadata() const override
        {
            return
                asf::Dictionary()
                    .insert("name", Model)
                    .insert("label", "BRL-CAD Object");
        }

        // Return metadata for the inputs of this object model.
        asf::DictionaryArray get_input_metadata() const override
        {
            asf::DictionaryArray metadata;

            metadata.push_back(
                asf::Dictionary()
                    .insert("name", "radius")
                    .insert("label", "Radius")
                    .insert("type", "numeric")
                    .insert("min",
                        asf::Dictionary()
                            .insert("value", "0.0")
                            .insert("type", "hard"))
                    .insert("max",
                        asf::Dictionary()
                            .insert("value", "10.0")
                            .insert("type", "soft"))
                    .insert("use", "optional")
                    .insert("default", "1.0"));

            return metadata;
        }

        // Create a new single empty object.
        asf::auto_release_ptr<asr::Object> create(
            const char*                 name,
            const asr::ParamArray&      params) const override
        {
            return asf::auto_release_ptr<asr::Object>(new BrlcadObject(name, params));
        }

        // Create objects, potentially from external assets.
        bool create(
            const char*                 name,
            const asr::ParamArray&      params,
            const asf::SearchPaths&     search_paths,
            const bool                  omit_loading_assets,
            asr::ObjectArray&           objects) const override
        {
            objects.push_back(create(name, params).release());
            return true;
        }
    };
}


//
// Plugin entry point.
//

extern "C"
{
    APPLESEED_DLL_EXPORT asr::IObjectFactory* appleseed_create_object_factory()
    {
        return new BrlcadObjectFactory();
    }
}
